#include "LiquidCrystal_I2C.h"
#include <unistd.h>

LiquidCrystal_I2C::LiquidCrystal_I2C(PropWare::I2CMaster* i2cbus, uint8_t lcd_addr, uint8_t lcd_cols, uint8_t lcd_rows, uint8_t charsize) {
	this->_addr = lcd_addr;
	this->_cols = lcd_cols;
	this->_rows = lcd_rows;
	this->_charsize = charsize;
	this->_backlightval = LCD_BACKLIGHT;
	this->_i2cbus = i2cbus;
}

void LiquidCrystal_I2C::begin() {
	this->_displayfunction = LCD_4BITMODE | LCD_1LINE | LCD_5x8DOTS;
	if (this->_rows > 1) this->_displayfunction |= LCD_2LINE;
	if ((this->_charsize != 0) && (this->_rows == 1)) this->_displayfunction |= LCD_5x10DOTS;

	LiquidCrystal_I2C::delay(50);

	expanderWrite(this->_backlightval);
	LiquidCrystal_I2C::delay(1000);

	write4bits(0x03 << 4);
	LiquidCrystal_I2C::delayMicro(4500);

	write4bits(0x03 << 4);
	LiquidCrystal_I2C::delayMicro(4500);

	write4bits(0x03 << 4);
	LiquidCrystal_I2C::delayMicro(150);

	write4bits(0x02 << 4);

	this->command(LCD_FUNCTIONSET | this->_displayfunction);

	this->_displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;
	this->display();

	this->clear();

	this->_displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;

	this->command(LCD_ENTRYMODESET | this->_displaymode);

	this->home();
}

void LiquidCrystal_I2C::clear() {
	this->command(LCD_CLEARDISPLAY);
	LiquidCrystal_I2C::delayMicro(2000);
}

void LiquidCrystal_I2C::home() {
	this->command(LCD_RETURNHOME);
	LiquidCrystal_I2C::delayMicro(2000);
}

void LiquidCrystal_I2C::noDisplay() {
	this->_displaycontrol &= ~LCD_DISPLAYON;
	this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

void LiquidCrystal_I2C::display() {
	this->_displaycontrol |= LCD_DISPLAYON;
	this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

void LiquidCrystal_I2C::noBlink() {
	this->_displaycontrol &= ~LCD_BLINKON;
	this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

void LiquidCrystal_I2C::blink() {
	this->_displaycontrol |= LCD_BLINKON;
	this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

void LiquidCrystal_I2C::noCursor() {
	this->_displaycontrol &= ~LCD_CURSORON;
	this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

void LiquidCrystal_I2C::cursor() {
	this->_displaycontrol |= LCD_CURSORON;
	this->command(LCD_DISPLAYCONTROL | this->_displaycontrol);
}

void LiquidCrystal_I2C::scrollDisplayLeft() { this->command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVELEFT); }

void LiquidCrystal_I2C::scrollDisplayRight() { this->command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVERIGHT); }

void LiquidCrystal_I2C::leftToRight() {
	this->_displaymode |= LCD_ENTRYLEFT;
	this->command(LCD_ENTRYMODESET | this->_displaymode);
}

void LiquidCrystal_I2C::rightToLeft() {
	this->_displaymode &= ~LCD_ENTRYLEFT;
	this->command(LCD_ENTRYMODESET | this->_displaymode);
}

void LiquidCrystal_I2C::noBacklight() {
	this->_backlightval = LCD_NOBACKLIGHT;
	this->expanderWrite(0);
}

void LiquidCrystal_I2C::backlight() {
	this->_backlightval = LCD_BACKLIGHT;
	this->expanderWrite(0);
}

bool LiquidCrystal_I2C::getBacklight() { return this->_backlightval == LCD_BACKLIGHT; }

void LiquidCrystal_I2C::autoscroll() {
	this->_displaymode |= LCD_ENTRYSHIFTINCREMENT;
	this->command(LCD_ENTRYMODESET | this->_displaymode);
}

void LiquidCrystal_I2C::noAutoscroll() {
	this->_displaymode &= ~LCD_ENTRYSHIFTINCREMENT;
	this->command(LCD_ENTRYMODESET | this->_displaymode);
}

void LiquidCrystal_I2C::createChar(uint8_t location, uint8_t charmap[]) {
	location &= 0x7; // we only have 8 locations 0-7
	this->command(LCD_SETCGRAMADDR | (location << 3));
	for (int i = 0; i < 8; i++) this->write(charmap[i]);
}

void LiquidCrystal_I2C::setCursor(uint8_t col, uint8_t row) {
	int row_offsets[] = {0x00, 0x40, 0x14, 0x54};
	if (row > this->_rows) row = this->_rows - 1;    // we count rows starting w/0
	this->command(LCD_SETDDRAMADDR | (col + row_offsets[row]));
}

void LiquidCrystal_I2C::command(uint8_t value) { this->send(value, 0); }

int LiquidCrystal_I2C::write(uint8_t value) {
	this->send(value, Rs);
	return 1;
}

void LiquidCrystal_I2C::blink_on() { this->blink(); }

void LiquidCrystal_I2C::blink_off() { this->noBlink(); }

void LiquidCrystal_I2C::cursor_on() { this->cursor(); }

void LiquidCrystal_I2C::cursor_off() { this->noCursor(); }

void LiquidCrystal_I2C::setBacklight(bool new_val) {
	if (new_val) this->backlight();		// turn backlight on
	else this->noBacklight();		// turn backlight off

	//new_val ? this->backlight() : this->noBacklight();
}

void LiquidCrystal_I2C::load_custom_character(uint8_t char_num, uint8_t *rows) { this->createChar(char_num, rows); }

void LiquidCrystal_I2C::printstr(const char c[]) { for (size_t i = 0; i < strlen(c); i++) this->write(c[i]); }

//------------------------------
void LiquidCrystal_I2C::send(uint8_t value, uint8_t mode) {
	uint8_t highnib = value & 0xf0;
	uint8_t lownib = (value << 4) & 0xf0;
	this->write4bits((highnib) | mode);
	this->write4bits((lownib) | mode);
}

void LiquidCrystal_I2C::write4bits(uint8_t value) {
	this->expanderWrite(value);
	this->pulseEnable(value);
}

void LiquidCrystal_I2C::expanderWrite(uint8_t _data) {
	/*i2c_start(this->_i2cbus);
	//address
	{
		uint8_t i2cAddr = this->_addr;
		i2cAddr <<= 1;
		i2cAddr &= -2;
		i2c_writeByte(this->_i2cbus, i2cAddr);
	}
	//data
	i2c_writeByte(this->_i2cbus, (int)(_data) | this->_backlightval);
	i2c_stop(this->_i2cbus);*/
	this->_i2cbus->put(this->_addr << 1, (int)(_data) | this->_backlightval);
}

void LiquidCrystal_I2C::pulseEnable(uint8_t _data) {
	this->expanderWrite(_data | En);
	LiquidCrystal_I2C::delayMicro(1);

	this->expanderWrite(_data & ~En);
	LiquidCrystal_I2C::delayMicro(50);
}
//------------------------------