#include "YM2612.h"

#include <time.h>
#include <sys/thread.h>
#include <unistd.h>
#include <propeller.h>
#include <cog.h>

YM2612::YM2612(
	PropWare::SimplePort *bus,
	PropWare::Pin::Mask CS,
	PropWare::Pin::Mask RD,
	PropWare::Pin::Mask WR,
	PropWare::Pin::Mask IC,
	PropWare::Pin::Mask A0,
	PropWare::Pin::Mask A1,
	PropWare::Pin::Mask RDY
) {
	this->_bus = bus;
	this->_CS = PropWare::Pin(CS, PropWare::Pin::Dir::OUT);
	this->_RD = PropWare::Pin(RD, PropWare::Pin::Dir::OUT);
	this->_WR = PropWare::Pin(WR, PropWare::Pin::Dir::OUT);
	this->_IC = PropWare::Pin(IC, PropWare::Pin::Dir::OUT);
	this->_A0 = PropWare::Pin(A0, PropWare::Pin::Dir::OUT);
	this->_A1 = PropWare::Pin(A1, PropWare::Pin::Dir::OUT);

	this->_RDY = PropWare::Pin(RDY);

	this->_CS.high();
	this->_RD.high();
	this->_WR.high();
	this->_IC.high();
	this->_A0.low();
	this->_A1.low();

	//this->_RDY.low();

	/*GPIOA->regs->CRH = (GPIOA->regs->CRH & 0x0FF00FFF) | 0x30033000; //PA11-YM_A0, PA12-YM_WR, PA15-YM_RD OUTPUT
	GPIOB->regs->CRL = (GPIOB->regs->CRL & 0xFFF00FF0) | 0x00033003; //PB0-YM_A1, PB3-YM_CS, PB4-YM_IC OUTPUT

	GPIOB->regs->BSRR = (1U << 3) << (16 * 0); //_CS HIGH
	GPIOA->regs->BSRR = (1U << 15) << (16 * 0); //_RD HIGH
	GPIOA->regs->BSRR = (1U << 12) << (16 * 0); //_WR HIGH
	GPIOB->regs->BSRR = (1U << 4) << (16 * 0); //_IC HIGH
	GPIOA->regs->BSRR = (1U << 11) << (16 * 1); //_A0 LOW
	GPIOB->regs->BSRR = (1U << 0) << (16 * 1); //_A1 LOW*/
}

void YM2612::Reset() {
    //GPIOB->regs->BSRR = (1U << 4) << (16 * 1); //_IC LOW
    //delayMicroseconds(25);
    //GPIOB->regs->BSRR = (1U << 4) << (16 * 0); //_IC HIGH

	//low(this->_IC);
	this->_IC.low();
	//usleep(25);
	//sleep37_5(25);
	//high(this->_IC);
	usleep(YM_RESET_WAIT);
	this->_IC.high();
}

void YM2612::Send(uint8_t addr, uint8_t data, bool setA1) { //0x52 = A1 LOW, 0x53 = A1 HIGH
	this->WaitReady();

    /*GPIOB->regs->BSRR = (1U << 0) << (16 * !setA1); //_A1 PB0
    GPIOA->regs->ODR &= ~(0x0800); //_A0 LOW
    GPIOB->regs->ODR &= ~(0x0808); //_CS LOW
    _bus->Write(addr);
    GPIOA->regs->ODR &= ~(0x1000); //_WR LOW
    GPIOA->regs->ODR |= 0x1000;    //_WR HIGH
    GPIOB->regs->ODR |= 0x0808;    //_CS HIGH
    GPIOA->regs->ODR |= 0x0800;    //_A0 HIGH
    GPIOB->regs->ODR &= ~(0x0808); //_CS LOW
    _bus->Write(data);
    GPIOA->regs->ODR &= ~(0x1000); //_WR LOW
    GPIOA->regs->ODR |= 0x1000;    //_WR HIGH
    GPIOB->regs->ODR |= 0x0808;    //_CS HIGH*/

#ifdef FLIPBUS
	addr = ~addr;
	data = ~data;
#endif

	if (setA1) this->_A1.high();
	else this->_A1.low();
	
	this->_A0.low();
	this->_CS.low();

	usleep(YM_WRITE_WAIT);
	
	this->_bus->write(addr);
	
	//usleep(1);
	//sleep37_5(2);
	usleep(YM_WRITE_WAIT);
	
	this->_WR.low();

	//usleep(5);
	//sleep37_5(5);
	usleep(YM_WRITE_WAIT);

	this->_WR.high();

	//usleep(5);
	usleep(YM_WRITE_WAIT);
	//sleep37_5(5);

	this->_CS.high();

	//usleep(5);
	//sleep37_5(5);
	usleep(YM_WRITE_WAIT);

	this->_A0.high();
	this->_CS.low();

	usleep(YM_WRITE_WAIT);

	this->_bus->write(data);

	//usleep(1);
	//sleep37_5(2);
	//usleep(YM_WRITE_WAIT);

	this->_WR.low();

	usleep(YM_WRITE_WAIT);
	//usleep(5);
	//sleep37_5(5);
	//usleep(YM_WRITE_WAIT);

	this->_WR.high();

	//usleep(5);
	//sleep37_5(5);
	usleep(YM_WRITE_WAIT);

	this->_CS.high();

	//usleep(5);
	//sleep37_5(5);
	usleep(YM_WRITE_WAIT);

	//A0 low
	//CS low
	//addr
	//delay 1 us
	//WR low
	//delay 5 us
	//WR high
	//delay 5 us
	//CS high

	//A0 high
	//CS low
	//data
	//delay 1 us
	//WR low
	//delay 5 us
	//WR high
	//delay 5 us
	//CS high

	/*
	if (setA1) high(this->_A1);
	else low(this->_A1);low(this->_A0);
	low(this->_CS);
	//OUTA |= addr << this->_D;
	//set_outputs(this->_D + 7, this->_D, addr);
	this->_bus->Write(addr);
	low(this->_WR);
	high(this->_WR);
	high(this->_CS);
	high(this->_A0);
	low(this->_CS);
	//OUTA |= data << this->_D;
	//set_outputs(this->_D + 7, this->_D, data);
	this->_bus->Write(data);
	low(this->_WR);
	high(this->_WR);
	high(this->_CS);*/

	//usleep(200);
}

void YM2612::SendPacked(uint16_t value, bool setA1){
	//uint8_t addr = (value & 0xFF00) >> 8;
	//uint8_t data = value & 0xFF;
	this->Send((uint8_t)((value & 0xFF00) >> 8), (uint8_t)(value & 0x00FF), setA1);
}

void YM2612::WaitReady() {
	//usleep(1);
	usleep(YM_WRITE_WAIT);
	//sleep37_5(2);
	this->_RDY.set_dir_in();
	//this->_RDY.set_dir(PropWare::Port::Dir::IN);
	this->_WR.high();
	this->_A1.low();
	this->_A0.low();
	this->_RD.low();
	//usleep(1);
	//sleep37_5(2);
	usleep(YM_WRITE_WAIT);
	this->_CS.low();
	this->_RDY.wait_until_low();
	this->_RD.high();
	this->_CS.high();
	//usleep(1);
	usleep(YM_WRITE_WAIT);
	//sleep37_5(2);
	this->_RDY.set_dir_out();
	//this->_RDY.set_dir(PropWare::Port::Dir::OUT);
}

void YM2612::sleep37_5(unsigned int n) {
	__napuntil(CNT + n);
}